#pragma once
/**
 * spline_interpolator.h - Utility classes for 1D cubic spline interpolation
 *
 * Copyright (C) 2017 University of Oxford
 */

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk

    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK


    LICENCE

    FMRIB Software Library, Release 6.0 (c) 2018, The University of
    Oxford (the "Software")

    The Software remains the property of the Oxford University Innovation
    ("the University").

    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.

    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.

    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.

    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Oxford
    University Innovation ("OUI"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    fsl@innovation.ox.ac.uk quoting Reference Project 9564, FSL.*/

#include <string>
#include <vector>

/**
 * Parameters for cubic spline starting at x=x0
 *
 * Equation is d(x-x0)^3 + c(x-x0)^2 + b(x-x0) + 1
 */
struct Spline
{
    Spline(double a = 0, double b = 0, double c = 0, double d = 0, double x0 = 0)
        : a(a)
        , b(b)
        , c(c)
        , d(d)
        , x0(x0)
    {
    }

    /** Evaluate the equation */
    double operator()(double x) const;

    /** Return human-readable equation to plug into FooPlot etc. */
    std::string eqn() const;

    double a, b, c, d, x0;
};

/**
 * Base class for anything which does spline interpolation
 */
class SplineInterpolator
{
public:
    /** Evaluate the interpolated function at an arbitrary x value */
    double operator()(double x) const;

protected:
    std::vector<Spline> m_splines;
};

/**
 * PCHIP (Piecewise Cubic Hermite Interpolating Polynomial) spline interpolator,
 * as used in Matlab version of the model
 *
 * The PCHIP method is designed to be shape-preserving at the expense of smoothness.
 * The second derivative is not necessarily continuous, however 'overshoots' are
 * avoided.
 */
class PchipInterpolator : public SplineInterpolator
{
public:
    /**
     * Generate a PCHIP interpolation with control points given by vectors of x
     * and y points
     */
    PchipInterpolator(std::vector<double> &x, std::vector<double> &y);

private:
    Spline get_spline(double x1, double y1, double m1, double x2, double y2, double m2);
    double edge_case(double h0, double h1, double m0, double m1);
    std::vector<double> get_derivatives(std::vector<double> &x, std::vector<double> &y);
};

/**
 * 'Natural' spline interpolator for comparison
 *
 * Natural splines ensure continuous second derivateives, which are set to zero at the
 * endpoints of the interval of interpolation.
 */
class NaturalSplineInterpolator : public SplineInterpolator
{
public:
    /**
     * Generate a natural spline interpolation with control points given by vectors of x
     * and y points
     */
    NaturalSplineInterpolator(std::vector<double> &x, std::vector<double> &y);
};
